package org.mortbay.jetty.client;
//========================================================================
//Copyright 2006-2007 Mort Bay Consulting Pty. Ltd.
//------------------------------------------------------------------------
//Licensed under the Apache License, Version 2.0 (the "License");
//you may not use this file except in compliance with the License.
//You may obtain a copy of the License at
//http://www.apache.org/licenses/LICENSE-2.0
//Unless required by applicable law or agreed to in writing, software
//distributed under the License is distributed on an "AS IS" BASIS,
//WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//See the License for the specific language governing permissions and
//limitations under the License.
//========================================================================

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.util.LinkedList;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

import org.mortbay.io.Buffer;
import org.mortbay.io.ByteArrayBuffer;

/**
*
* @author Greg Wilkins
* @author Guillaume Nodet
*/
class HttpDestination
{
    private InetSocketAddress _address;
    private LinkedList<HttpConnection> _connections=new LinkedList<HttpConnection>();
    private LinkedList<HttpConnection> _idle=new LinkedList<HttpConnection>();
    private HttpClient _client;
    private boolean _ssl;
    private int _maxConnections;
    private AtomicInteger _pendingConnections = new AtomicInteger(0);
    
    /* The queue of exchanged for this destination if connections are limited */
    private LinkedList<HttpExchange> _exchanges=new LinkedList<HttpExchange>();

    /* ------------------------------------------------------------ */
    HttpDestination(HttpClient pool, InetSocketAddress address, boolean ssl, int maxConnections)
    {
        _client=pool;
        _address=address;
        _ssl=ssl;
        _maxConnections=maxConnections;
        String host = address.getHostName();
        if (address.getPort() != (_ssl ? 443 : 80)) {
            host += ":" + address.getPort();
        }
    }

    /* ------------------------------------------------------------ */
    public InetSocketAddress getAddress()
    {
        return _address;
    }

    /* ------------------------------------------------------------------------------- */
    public HttpConnection getConnection() throws IOException
    {
        synchronized (this)
        {
            // Find an idle connection
            if (_idle.size() > 0)
            {
                HttpConnection connection = _idle.removeFirst();
                if (connection.getEndPoint().isOpen())
                {
                    System.err.println("use idle "+connection);
                    return connection;
                }
                else
                {
                    _connections.remove(connection);
                }
            }

            if (_connections.size()+_pendingConnections.get() <_maxConnections)
            {
                HttpConnection connection=startNewConnection();
                return connection;
            }
            return null;
        }
    }

    /* ------------------------------------------------------------------------------- */
    protected HttpConnection startNewConnection() throws UnknownHostException, IOException
    {
        _pendingConnections.incrementAndGet();
        return _client._connector.startConnection(this);
    }

    /* ------------------------------------------------------------------------------- */
    public void onNewConnection(HttpConnection connection) throws IOException
    {
        System.err.println("onNewConnection "+connection);
        _pendingConnections.decrementAndGet();
        synchronized (this)
        {
            _connections.add(connection);
            if (_exchanges.size()==0)
                _idle.add(connection);
            else
                connection.send(_exchanges.removeFirst());
        }
    }

    /* ------------------------------------------------------------------------------- */
    public void returnConnection(HttpConnection connection, boolean close) throws IOException
    {
        System.err.println("returnConnection "+connection+" to "+this);
        if (!close && connection.getEndPoint().isOpen())
        {
            synchronized (this)
            {
                _connections.add(connection);
                if (_exchanges.size()==0)
                {
                    System.err.println("idle "+connection+" in "+this);
                    _idle.add(connection);
                }
                else
                    connection.send(_exchanges.removeFirst());
            }
        }
        else 
        {
            _connections.remove(connection);
            synchronized (this)
            {
                if (!_exchanges.isEmpty())
                    startNewConnection();
            }
        }
    }
    
    /* ------------------------------------------------------------ */
    public void send(HttpExchange ex) throws IOException
    {
        synchronized(this)
        {
            HttpConnection connection=getConnection();
            if (connection!=null)
                connection.send(ex);
            else
            {
                System.err.println("queue "+ex);
                _exchanges.add(ex);
            }
        }
    }

    /* ------------------------------------------------------------ */
    public synchronized String toString()
    {
        return "HttpDestination@"+hashCode()+"//"+_address.getHostName()+":"+_address.getPort()+"("+_connections.size()+","+_idle.size()+","+_exchanges.size()+")";
    }
    
}